@@ -1,19 +1,18 @@
-from ryu.base import app_manager
 from ryu.controller import ofp_event
-from ryu.controller.handler import MAIN_DISPATCHER
+from ryu.controller.handler import MAIN_DISPATCHER, CONFIG_DISPATCHER
 from ryu.controller.handler import set_ev_cls
-from ryu.ofproto import ofproto_v1_0
 from ryu.lib.mac import haddr_to_bin
 from ryu.lib.packet import packet
 from ryu.lib.packet import ethernet
 from ryu.lib.packet import ether_types
 from csle_ryu.dao.ryu_controller_type import RYUControllerType
 from csle_ryu.monitor.flow_and_port_stats_monitor import FlowAndPortStatsMonitor
+import csle_ryu.constants.constants as constants
 
 
 class LearningSwitchController(FlowAndPortStatsMonitor):
     """
-    RYU Controller implementing a learning L2 switch
+    RYU Controller implementing a learning L2 switch for OpenFlow 1.3
     """
 
     def __init__(self, *args, **kwargs):
@@ -31,30 +30,68 @@ class LearningSwitchController(FlowAndPortStatsMonitor):
         # Controller type
         self.controller_type = RYUControllerType.LEARNING_SWITCH
 
-    def add_flow(self, datapath, in_port, dst, src, actions) -> None:
+
+    @set_ev_cls(ofp_event.EventOFPSwitchFeatures, CONFIG_DISPATCHER)
+    def switch_features_handler(self, ev):
+        """
+        Handler called after OpenFlow handshake with switch completed. It adds teh Table-miss flow entry to
+        the flow tables of the switch so that next packet which yield a flow-table-miss will be sent to the
+        controller
+
+        :param ev: the handshake complete event
+        :return:
+        """
+        # the datapath, i.e. abstraction of the link to the switch
+        datapath = ev.msg.datapath
+
+        self.logger.info(f"[SDN-Controller {self.controller_type}] OpenFlow handshake with switch DPID:{datapath.id}, "
+                         f"completed, adding the default flow-miss entry to the table")
+
+        # Extract version and packet parser
+        openflow_protocol = datapath.ofproto
+        parser = datapath.ofproto_parser
+
+        # Configure the table-miss flow to install at the table
+
+        # First, setup the matcher that will match packets to flows based on the headers
+        # It should match any packet
+        match = parser.OFPMatch()
+
+        # Second, define action when flows are matched
+        actions = [parser.OFPActionOutput(openflow_protocol.OFPP_CONTROLLER, constants.RYU.PACKET_BUFFER_MAX_LEN)]
+
+        # Third, define the priority, we set the lowest priority so that this flow rule is matched only if no other
+        # flow is matched
+        priority = 0
+
+        # Send request to the switch to add the flow
+        self.add_flow(datapath, priority, match, actions)
+
+    def add_flow(self, datapath, priority, match, actions, buffer_id=None) -> None:
         """
         Utility method for adding a flow to a switch
 
         :param datapath: the datapath, i.e. abstraction of the link to the switch
-        :param in_port: the incoming port for matching the flow
-        :param dst: the destination address for matching the flow
-        :param src: the source address for matching the flow
+        :param priority: the priority of the flow (higher priority are prioritized)
+        :param match: the pattern for matching packets to the flow based on the header
         :param actions: the actions to take when the flow is matched, e.g. where to send the packet
+        :param buffer_id: the id of the buffer where packets for this flow are queued if they cannot be sent
         :return: None
         """
         openflow_protocol = datapath.ofproto # Extract the openflow protocol used for communicating with the switch
+        parser = datapath.ofproto_parser # extract packet parser
 
-        # Create a match pattern that the switch should use to match incoming packet headers to flows.
-        # The match pattern includes the incoming port, the destination address (MAC or IP),
-        # and the source address (MAC or IP)
-        match = datapath.ofproto_parser.OFPMatch(in_port=in_port, dl_dst=haddr_to_bin(dst), dl_src=haddr_to_bin(src))
+        # Define the instruction that the switch should perform if the flow is matched
+        instruction = [parser.OFPInstructionActions(openflow_protocol.OFPIT_APPLY_ACTIONS, actions)]
 
         # Create an OpenFlow Modify-State message to add a new flow
-        mod = datapath.ofproto_parser.OFPFlowMod(
-            datapath=datapath, match=match, cookie=0,
-            command=openflow_protocol.OFPFC_ADD, idle_timeout=0, hard_timeout=0,
-            priority=openflow_protocol.OFP_DEFAULT_PRIORITY,
-            flags=openflow_protocol.OFPFF_SEND_FLOW_REM, actions=actions)
+        if buffer_id is not None:
+            mod = parser.OFPFlowMod(datapath=datapath, buffer_id=buffer_id,
+                                    priority=priority, match=match,
+                                    instructions=instruction)
+        else:
+            mod = parser.OFPFlowMod(datapath=datapath, priority=priority,
+                                    match=match, instructions=instruction)
 
         # Send the message to the switch
         datapath.send_msg(mod)
@@ -72,24 +109,34 @@ class LearningSwitchController(FlowAndPortStatsMonitor):
         # Extract metadata from the event received from the switch
         msg = ev.msg
         datapath = msg.datapath
+
+        # message protocol and parser and input port
         openflow_protocol = datapath.ofproto
-        pkt = packet.Packet(msg.data) # Create a Packet object from the event data
-        eth = pkt.get_protocol(ethernet.ethernet)
+        parser = datapath.ofproto_parser
+        in_port = msg.match['in_port']
+
+        pkt = packet.Packet(msg.data)
+        eth = pkt.get_protocols(ethernet.ethernet)[0]
+
         if eth.ethertype == ether_types.ETH_TYPE_LLDP:
             # ignore link-layer discovery protocol (lldp) packet
             return
+
         dst_mac_address = eth.dst
         src_mac_address = eth.src
+
         # Each switch is identified by a datapath 64-bit ID (DPID) containing the 48-bit MAC address
         # as well as 16-bit additional ID bits
-        datapath_id = datapath.id
+        datapath_id = format(datapath.id, "d").zfill(16)
+
+        # Initialize empty row in the mac-to-port table
         self.mac_to_port.setdefault(datapath_id, {})
 
         self.logger.info(f"[SDN-Controller {self.controller_type}] received packet in, DPID:{datapath_id}, "
-                     f"src_mac_address:{src_mac_address}, dst_mac_address:{dst_mac_address}, port number:{msg.in_port}")
+                     f"src_mac_address:{src_mac_address}, dst_mac_address:{dst_mac_address}, port number:{in_port}")
 
         # learn a mac address to avoid FLOOD next time, i.e. map the port of the switch to the source MAC address
-        self.mac_to_port[datapath_id][src_mac_address] = msg.in_port
+        self.mac_to_port[datapath_id][src_mac_address] = in_port
 
         if dst_mac_address in self.mac_to_port[datapath_id]:
             # If the destination MAC address already exists in the table, lookup the corresponding port
@@ -99,7 +146,7 @@ class LearningSwitchController(FlowAndPortStatsMonitor):
             out_port = openflow_protocol.OFPP_FLOOD
 
         # Prepare the actions to be sent to the switch
-        actions = [datapath.ofproto_parser.OFPActionOutput(out_port)]
+        actions = [parser.OFPActionOutput(out_port)]
 
         # If the destination MAC already has been learned, install a flow in the switch that tells the switch
         # which outgoing port it should use whenever it receives frames that matches the given source port, source MAC,
@@ -107,7 +154,16 @@ class LearningSwitchController(FlowAndPortStatsMonitor):
         # the switch does not need to flood the LAN, it can just lookup the correct destination port from the
         # flow table.
         if out_port != openflow_protocol.OFPP_FLOOD:
-            self.add_flow(datapath, msg.in_port, dst_mac_address, src_mac_address, actions)
+            # Create a matcher object which defines how packets are matched to the flow
+            match = parser.OFPMatch(in_port=in_port, eth_dst=dst_mac_address, eth_src=src_mac_address)
+
+            # verify if we have a valid buffer_id, if yes then we dont have to assign a new buffer id and
+            # we dont have to send back the packet data since it is already queued in the buffer
+            if msg.buffer_id != openflow_protocol.OFP_NO_BUFFER:
+                self.add_flow(datapath, 1, match, actions, msg.buffer_id)
+                return
+            else:
+                self.add_flow(datapath, 1, match, actions)
 
         # The switch normally keeps packet it does not know how to forward in a queue/buffer. We extract the
         # buffer id from the openflow event. However if the switch does not use buffering, it will send the complete
@@ -119,38 +175,8 @@ class LearningSwitchController(FlowAndPortStatsMonitor):
 
         # Prepare OFP message to send to the switch and instruct it what to do with this packet.
         out = datapath.ofproto_parser.OFPPacketOut(
-            datapath=datapath, buffer_id=msg.buffer_id, in_port=msg.in_port,
+            datapath=datapath, buffer_id=msg.buffer_id, in_port=in_port,
             actions=actions, data=data)
 
         # Send the message to the Switch
         datapath.send_msg(out)
-
-    @set_ev_cls(ofp_event.EventOFPPortStatus, MAIN_DISPATCHER)
-    def _port_status_handler(self, ev) -> None:
-        """
-        Handler called when the controller receives a message from the switch which indicates that one of its
-        ports has changed
-
-        :param ev: the port-change evnet
-        :return: None
-        """
-
-        # Extract metadata from the received OpenFlow event
-        msg = ev.msg
-        reason = msg.reason
-        port_no = msg.desc.port_no
-        openflow_protocol = msg.datapath.ofproto
-        datapath = msg.datapath
-        datapath_id = datapath.id
-
-        if reason == openflow_protocol.OFPPR_ADD:
-            self.logger.info(f"[SDN-Controller {self.controller_type}] a port {port_no} was added to switch with "
-                         f"DPID: {datapath_id}")
-        elif reason == openflow_protocol.OFPPR_DELETE:
-            self.logger.info(f"[SDN-Controller {self.controller_type}] port {port_no} was removed from switch with "
-                         f"DPID: {datapath_id}")
-        elif reason == openflow_protocol.OFPPR_MODIFY:
-            self.logger.info(f"[SDN-Controller {self.controller_type}] port {port_no} on switch with "
-                         f"DPID: {datapath_id} was modified")
-        else:
-            self.logger.info(f"[SDN-Controller {self.controller_type}] Illegal port state: {port_no}, {reason}")
\ No newline at end of file
